import collections
import concurrent.futures
import time
import types

from . import packlib
from . import poollib
from . import server_socket
from . import thread


class Server:

  def __init__(self, port, name='Server', workers=1, errors=True, **kwargs):
    self.socket = server_socket.ServerSocket(port, name, **kwargs)
    self.loop = thread.Thread(self._loop, name=f'{name}Loop')
    self.methods = {}
    self.jobs = set()
    self.workers = workers
    self.errors = errors
    self.running = False
    self.pool = poollib.ThreadPool(workers, 'default_pool')
    self.postfn_pool = poollib.ThreadPool(1, 'postfn')
    self.postfn_inp = collections.deque()
    self.postfn_out = collections.deque()
    self.metrics = dict(send=0, recv=0, time=time.time())
    self.pools = [self.pool, self.postfn_pool]

  def bind(self, name, workfn, postfn=None, workers=0):
    assert not self.running
    assert name not in self.methods, name
    if workers:
      pool = poollib.ThreadPool(workers, '{name}_pool')
      self.pools.append(pool)
    else:
      pool = self.pool
    requests = collections.deque()
    available = (workers or self.workers) + 1
    self.methods[name] = types.SimpleNamespace(
        workfn=workfn, postfn=postfn, pool=pool,
        requests=requests, available=available)

  def start(self, block=True):
    assert not self.running
    self.running = True
    self.loop.start()
    if block:
      self.loop.join(timeout=None)

  def close(self, timeout=None, internal=False):
    assert self.running
    self.socket.shutdown()
    self.running = False
    if not internal:
      self.loop.join(timeout)
      self.loop.kill()
    [x.close() for x in self.pools]
    self.socket.close()

  def stats(self):
    now = time.time()
    mets = self.metrics
    self.metrics = dict(send=0, recv=0, time=now)
    dur = now - mets['time']
    stats = {
        'numsend': mets['send'],
        'numrecv': mets['recv'],
        'sendrate': mets['send'] / dur,
        'recvrate': mets['recv'] / dur,
        'requests': sum(len(m.requests) for m in self.methods.values()),
        'jobs': len(self.jobs),
    }
    if any(method.postfn for method in self.methods.values()):
      stats.update({
          'post_iqueue': len(self.postfn_inp),
          'post_oqueue': len(self.postfn_out),
      })
    return stats

  def __enter__(self):
    self.start(block=False)
    return self

  def __exit__(self, *e):
    self.close()

  def _loop(self):
    methods = list(self.methods.values())
    pending = 0
    while self.running or pending:
      while True:  # Loop syntax used to break on error.
        if not self.running:  # Do not accept further requests.
          break
        try:
          addr, data = self.socket.recv(timeout=0.0001)
        except TimeoutError:
          break
        if len(data) < 8:
          self._error(addr, bytes(8), 1, 'Message too short')
          break
        reqnum, data = bytes(data[:8]), data[8:]
        try:
          strlen = int.from_bytes(data[:8], 'little', signed=False)
          name = bytes(data[8: 8 + strlen]).decode('utf-8')
          data = packlib.unpack(data[8 + strlen:])
        except Exception:
          self._error(addr, reqnum, 2, 'Could not decode message')
          break
        if name not in self.methods:
          self._error(addr, reqnum, 3, f'Unknown method {name}')
          break
        self.metrics['recv'] += 1
        method = self.methods[name]
        method.requests.append((addr, reqnum, data))
        pending += 1
        break  # We do not actually want to loop.

      for method in methods:
        if method.requests and method.available:
          method.available -= 1
          addr, reqnum, data = method.requests.popleft()
          job = method.pool.submit(method.workfn, *data)
          job.method = method
          job.addr = addr
          job.reqnum = reqnum
          self.jobs.add(job)
          if method.postfn:
            self.postfn_inp.append(job)

      completed, self.jobs = concurrent.futures.wait(
          self.jobs, 0.0001, concurrent.futures.FIRST_COMPLETED)
      for job in completed:
        try:
          data = job.result()
          if job.method.postfn:
            data, _ = data
          data = packlib.pack(data)
          status = int(0).to_bytes(8, 'little', signed=False)
          self.socket.send(job.addr, job.reqnum, status, *data)
          self.metrics['send'] += 1
        except Exception as e:
          self._error(job.addr, job.reqnum, 4, f'Error in server method: {e}')
        finally:
          if not job.method.postfn:
            job.method.available += 1
            pending -= 1

      if completed:
        # Call postfns in the order the requests were received.
        while self.postfn_inp and self.postfn_inp[0].done():
          job = self.postfn_inp.popleft()
          _, info = job.result()
          postjob = self.postfn_pool.submit(job.method.postfn, info)
          postjob.method = job.method
          self.postfn_out.append(postjob)

      while self.postfn_out and self.postfn_out[0].done():
        postjob = self.postfn_out.popleft()
        postjob.result()  # Check if there was an error.
        postjob.method.available += 1
        pending -= 1

  def _error(self, addr, reqnum, status, message):
    status = status.to_bytes(8, 'little', signed=False)
    data = message.encode('utf-8')
    self.socket.send(addr, reqnum, status, data)
    if self.errors:
      # Wait until the error is delivered to the client and then raise.
      self.close(internal=True)
      raise RuntimeError(message)



import time

import pytest
import portal


class TestSocket:

  @pytest.mark.parametrize('ipv6', (False, True))
  def test_basic(self, ipv6):
    port = portal.free_port()
    server = portal.ServerSocket(port, ipv6=ipv6)
    client = portal.ClientSocket(port, ipv6=ipv6)
    client.send(b'foo')
    addr, data = server.recv()
    assert addr[0] == '::1' if ipv6 else '127.0.0.1'
    assert data == b'foo'
    server.send(addr, b'bar')
    assert client.recv() == b'bar'
    server.close()
    client.close()

  def test_multi_buffer(self):
    port = portal.free_port()
    server = portal.ServerSocket(port)
    client = portal.ClientSocket(port)
    client.send(b'foo', b'bar', b'baz')
    addr, data = server.recv()
    assert data == b'foobarbaz'
    server.send(addr, b'ab', b'c')
    assert client.recv() == b'abc'
    server.close()
    client.close()

  def test_multiple_send(self):
    port = portal.free_port()
    server = portal.ServerSocket(port)
    client = portal.ClientSocket(port)
    client.send(b'foo')
    client.send(b'ba', b'r')
    client.send(b'baz')
    assert server.recv()[1] == b'foo'
    assert server.recv()[1] == b'bar'
    assert server.recv()[1] == b'baz'
    assert len(server.connections) == 1
    addr = server.connections[0]
    server.send(addr, b'baz')
    server.send(addr, b'ba', b'r')
    server.send(addr, b'foo')
    assert client.recv() == b'baz'
    assert client.recv() == b'bar'
    assert client.recv() == b'foo'
    server.close()
    client.close()

  @pytest.mark.parametrize('repeat', range(3))
  def test_disconnect_server(self, repeat):
    port = portal.free_port()
    server = portal.ServerSocket(port)
    client = portal.ClientSocket(port, autoconn=False)
    client.connect()
    server.close()
    with pytest.raises(portal.Disconnected):
      client.recv()
    server = portal.ServerSocket(port)
    with pytest.raises(portal.Disconnected):
      client.recv()
    client.connect()
    time.sleep(0.2)
    server.send(server.connections[0], b'foo')
    assert client.recv() == b'foo'
    server.close()
    time.sleep(0.2)
    with pytest.raises(portal.Disconnected):
      client.send(b'bar')

  @pytest.mark.parametrize('repeat', range(5))
  def test_disconnect_client(self, repeat):
    port = portal.free_port()
    server = portal.ServerSocket(port)
    client = portal.ClientSocket(port)
    client.send(b'foo')
    assert server.recv()[1] == b'foo'
    assert len(server.connections) == 1
    client.close()
    time.sleep(0.2)
    assert len(server.connections) == 0
    client = portal.ClientSocket(port)
    time.sleep(0.2)
    assert len(server.connections) == 1
    server.close()
    client.close()

  @pytest.mark.parametrize('repeat', range(3))
  def test_server_dies(self, repeat):
    port = portal.free_port()
    q = portal.context.mp.Queue()

    def server_fn(port, q):
      # Receive exactly one message and then exit wihout close().
      server = portal.ServerSocket(port)
      x = bytes(server.recv()[1])
      q.put(x)
      q.close()
      q.join_thread()

    def client_fn(port, q):
      client = portal.ClientSocket(
          port,
          autoconn=False,
          keepalive_after=1,
          keepalive_every=1,
          keepalive_fails=1)
      client.connect()
      try:
        assert client.connected
        while True:
          client.send(b'method')
          time.sleep(0.1)
      except portal.Disconnected:
        q.put(b'bye')
        client.connect(timeout=None)
        client.send(b'hi')
      q.close()
      q.join_thread()
      client.close()

    server = portal.Process(server_fn, port, q, start=True)
    client = portal.Process(client_fn, port, q, start=True)
    assert q.get() == b'method'
    server.join()
    assert q.get() == b'bye'
    server = portal.Process(server_fn, port, q, start=True)
    server.join()
    client.join()
    assert q.get() == b'hi'

  @pytest.mark.parametrize('repeat', range(3))
  def test_twoway(self, repeat, size=1024 ** 2, prefetch=8):

    def server(port):
      server = portal.ServerSocket(port)
      expected = bytearray(size)
      while True:
        addr, data = server.recv()
        if data == b'exit':
          server.send(addr, b'exit')
          break
        server.send(addr, data)
        assert len(data) == size
        assert data == expected
      server.close()

    def client(port):
      data = bytearray(size)
      client = portal.ClientSocket(port)
      for _ in range(prefetch):
        client.send(data)
      for _ in range(100):
        client.send(data)
        result = client.recv()
        assert len(result) == size
      client.send(b'exit')
      while client.recv() != b'exit':
        pass
      client.close()

    port = portal.free_port()
    portal.run([
        portal.Process(server, port),
        portal.Process(client, port),
    ])

  @pytest.mark.parametrize('repeat', range(3))
  def test_shutdown(self, repeat):

   def server(port):
     server = portal.ServerSocket(port)
     addr, data = server.recv()
     assert data == b'foo'
     large_result = bytes(1024 ** 2)
     server.send(addr, large_result)
     server.close()

   def client(port):
     client = portal.ClientSocket(port)
     client.send(b'foo')
     assert client.recv() == bytes(1024 ** 2)
     client.close()

   port = portal.free_port()
   portal.run([
       portal.Process(server, port),
       portal.Process(client, port),
   ])



import os
import pathlib
import time

import pytest
import portal


class TestErrfile:

  def test_file(self, tmpdir):
    errfile = pathlib.Path(tmpdir) / 'error'

    def fn():
      portal.setup(errfile=errfile, interval=0.1)
      try:
        raise ValueError
      except Exception as e:
        portal.context.error(e, 'worker')

    portal.Process(fn, start=True).join()
    content = errfile.read_text()
    assert "Error in 'worker' (ValueError):" == content.split('\n')[0]
    assert 'Traceback (most recent call last)' in content
    assert 'line' in content
    assert 'in fn' in content

  @pytest.mark.parametrize('repeat', range(5))
  def test_sibling_procs(self, tmpdir, repeat):
    errfile = pathlib.Path(tmpdir) / 'error'
    barrier = portal.context.mp.Barrier(3)

    def fn1(barrier, errfile):
      portal.setup(errfile=errfile, interval=0.1)
      barrier.wait()
      raise ValueError('reason')

    def fn2(barrier, errfile):
      portal.setup(errfile=errfile, interval=0.1)
      barrier.wait()
      while True:
        time.sleep(0.1)

    worker1 = portal.Process(fn1, barrier, errfile, start=True)
    worker2 = portal.Process(fn2, barrier, errfile, start=True)
    barrier.wait()
    worker1.join()
    worker2.join()
    content = errfile.read_text()
    first_line = content.split('\n')[0]
    assert "Error in 'fn1' (ValueError: reason):" == first_line
    assert not worker1.running
    assert not worker2.running
    # The first worker may shut itself down or be shut down based on its own
    # error file watcher, based on how the threads context-switch.
    assert worker1.exitcode in (1, 2)
    assert worker2.exitcode == 2

  @pytest.mark.parametrize('repeat', range(3))
  def test_nested_procs(self, tmpdir, repeat):
    errfile = pathlib.Path(tmpdir) / 'error'
    ready = portal.context.mp.Barrier(7)
    queue = portal.context.mp.Queue()

    def outer(ready, queue, errfile):
      portal.setup(errfile=errfile, interval=0.1)
      portal.Process(inner, ready, queue, name='inner', start=True)
      portal.Thread(hang_thread, ready, start=True)
      portal.Process(hang_process, ready, queue, start=True)
      queue.put(os.getpid())
      queue.close()
      queue.join_thread()
      ready.wait()  # 1
      while True:
        time.sleep(0.1)

    def inner(ready, queue):
      assert portal.context.errfile
      portal.Thread(hang_thread, ready, start=True)
      portal.Process(hang_process, ready, queue, start=True)
      queue.put(os.getpid())
      queue.close()
      queue.join_thread()
      ready.wait()  # 2
      raise ValueError('reason')

    def hang_thread(ready):
      ready.wait()  # 3, 4
      while True:
        time.sleep(0.1)

    def hang_process(ready, queue):
      assert portal.context.errfile
      queue.put(os.getpid())
      queue.close()
      queue.join_thread()
      ready.wait()  # 5, 6
      while True:
        time.sleep(0.1)

    worker = portal.Process(
        outer, ready, queue, errfile, name='outer', start=True)
    ready.wait()  # 7
    worker.join()
    content = errfile.read_text()
    assert "Error in 'inner' (ValueError: reason):" == content.split('\n')[0]
    assert not worker.running
    pids = [queue.get() for _ in range(4)]
    time.sleep(2.0)  # On some systems this can take a while.
    assert not portal.proc_alive(pids[0])
    assert not portal.proc_alive(pids[1])
    assert not portal.proc_alive(pids[2])
    assert not portal.proc_alive(pids[3])



import atexit
import errno
import traceback

import cloudpickle
import psutil

from . import contextlib
from . import utils


class Process:

  """
  This process implementation extends the standard Python process as follows:

  1. It provides a kill() method that not only terminates the process itself
  but also all its nested child processes. This prevents subprocesses from
  lingering in the background after the Python program has ended.

  2. The process terminates its nested child processes when it encounters an
  error. This prevents lingering subprocesses on error.

  3. When the parent process encounters an error, its subprocess will be killed
  via `atexit`, preventing hangs.

  4. It inherits the context object of its parent process, which provides
  cloudpickled initializer functions for each nested child process and error
  file watching for global shutdown on error.

  5. Standard Python subprocesses do not reliably return the running() state of
  the process. This class makes it more reliable.
  """

  def __init__(self, fn, *args, name=None, start=False):
    name = name or getattr(fn, '__name__', 'process')
    fn = cloudpickle.dumps(fn)
    options = contextlib.context.options()
    self.ready = contextlib.context.mp.Barrier(2)
    self.process = contextlib.context.mp.Process(
        target=self._wrapper, name=name,
        args=(options, self.ready, name, fn, args))
    self.started = False
    self.killed = False
    self.thepid = None
    atexit.register(self.kill)
    contextlib.context.add_worker(self)
    start and self.start()

  @property
  def name(self):
    return self.process.name

  @property
  def pid(self):
    return self.thepid

  @property
  def running(self):
    if not self.started:
      return False
    if not self.process.is_alive():
      return False
    return utils.proc_alive(self.pid)

  @property
  def exitcode(self):
    exitcode = self.process.exitcode
    if self.killed and exitcode is None:
      return -9
    return exitcode

  def start(self):
    assert not self.started
    self.started = True
    self.process.start()
    self.ready.wait()
    self.thepid = self.process.pid
    assert self.thepid is not None
    return self

  def join(self, timeout=None):
    assert self.started
    self.process.join(timeout)
    return self

  def kill(self, timeout=1):
    assert self.started
    try:
      children = list(psutil.Process(self.pid).children(recursive=True))
    except psutil.NoSuchProcess:
      children = []
    try:
      self.process.terminate()
      self.process.join(timeout)
      self.process.kill()
      utils.kill_proc(children, timeout)
    except OSError as e:
      if e.errno != errno.ESRCH:
        contextlib.context.error(e, self.name)
        contextlib.context.shutdown(exitcode=1)
    self.killed = True
    return self

  def __repr__(self):
    attrs = ('name', 'pid', 'running', 'exitcode')
    attrs = [f'{k}={getattr(self, k)}' for k in attrs]
    return 'Process(' + ', '.join(attrs) + ')'

  @staticmethod
  def _wrapper(options, ready, name, fn, args):
    exitcode = 0
    try:
      ready.wait()
      contextlib.setup(**options)
      fn = cloudpickle.loads(fn)
      exitcode = fn(*args)
      exitcode = exitcode if isinstance(exitcode, int) else 0
    except (SystemExit, KeyboardInterrupt) as e:
      compact = traceback.format_tb(e.__traceback__)
      compact = '\n'.join([line.split('\n', 1)[0] for line in compact])
      print(f"Killed process '{name}' at:\n{compact}")
      exitcode = 2
    except Exception as e:
      contextlib.context.error(e, name)
      exitcode = 1
    finally:
      contextlib.context.shutdown(exitcode)



import threading

import numpy as np
import portal

from . import client
from . import packlib
from . import process
from . import server
from . import server_socket
from . import sharray
from . import thread
from . import utils


class BatchServer:

  def __init__(
      self, port, name='Server', workers=1, errors=True,
      process=True, shmem=False, **kwargs):
    inner_port = utils.free_port()
    assert port != inner_port, (port, inner_port)
    self.name = name
    self.server = server.Server(inner_port, name, workers, errors, **kwargs)
    if process:
      self.running = portal.context.mp.Event()
    else:
      self.running = threading.Event()
    self.process = process
    self.batsizes = {}
    self.batargs = (
        self.running, port, inner_port, f'{name}Batcher',
        self.batsizes, errors, shmem, kwargs)
    self.started = False

  def bind(self, name, workfn, donefn=None, batch=0, workers=0):
    assert not self.started
    self.batsizes[name] = batch
    self.server.bind(name, workfn, donefn, workers=workers)

  def start(self, block=True):
    assert not self.started
    self.started = True
    self.running.set()
    if self.process:
      self.batcher = process.Process(
          batcher, *self.batargs, name=f'{self.name}Batcher', start=True)
    else:
      self.batcher = thread.Thread(
          batcher, *self.batargs, name=f'{self.name}Batcher', start=True)
    self.server.start(block=block)

  def close(self, timeout=None):
    assert self.started
    self.running.clear()
    self.server.close(timeout)
    self.batcher.join(timeout)
    self.batcher.kill()

  def stats(self):
    return self.server.stats()

  def __enter__(self):
    self.start(block=False)
    return self

  def __exit__(self, *e):
    self.close()


def batcher(
    running, outer_port, inner_port, name, batsizes, errors, shmem,
    kwargs):

  def maybe_recv(outer, inner, jobs, batches):
    if not running.is_set():  # Do not accept further requests.
      return
    try:
      addr, data = outer.recv(timeout=0.0001)
    except TimeoutError:
      return
    reqnum = bytes(data[:8])
    data = data[8:]
    strlen = int.from_bytes(data[:8], 'little', signed=False)
    data = data[8:]
    name, data = bytes(data[:strlen]).decode('utf-8'), data[strlen:]
    if name not in batsizes:
      send_error(addr, reqnum, 3, f'Unknown method {name}')
      return
    data = packlib.unpack(data)
    batch_size = batsizes[name]
    if not batch_size:
      job = inner.call(name, *data)
      job.args = (False, addr, reqnum)
      jobs.append(job)
      return
    leaves, structure = packlib.tree_flatten(data)
    leaves = [np.asarray(x) for x in leaves]
    if any(x.dtype == object for x in leaves):
      send_error(addr, reqnum, 5, 'Only array arguments can be batched.')
      return
    if name not in batches:
      if shmem:
        buffers = [
            sharray.SharedArray((batch_size, *leaf.shape), leaf.dtype)
            for leaf in leaves]
      else:
        buffers = [
            np.empty((batch_size, *leaf.shape), leaf.dtype)
            for leaf in leaves]
      batches[name] = ([], [], structure, buffers)
    addrs, reqnums, reference, buffers = batches[name]
    if structure != reference:
      send_error(addr, reqnum, 6, (
          f'Argument structure {structure} does not match previous ' +
          f'requests with structure {reference} for batched server ' +
          f'method {name}.'))
      return
    index = len(addrs)
    addrs.append(addr)
    reqnums.append(reqnum)
    for buffer, leaf in zip(buffers, leaves):
      buffer[index] = leaf
    if len(addrs) == batch_size:
      del batches[name]
      data = packlib.tree_unflatten(buffers, reference)
      job = inner.call(name, *data)
      job.args = (True, addrs, reqnums)
      jobs.append(job)

  def maybe_send(outer, inner, jobs):
    done, waiting = [], []
    [done.append(x) if x.done() else waiting.append(x) for x in jobs]
    for job in done:
      batched, addr, reqnum = job.args
      try:
        result = job.result()
      except RuntimeError as e:
        if batched:
          for i, (addr, reqnum) in enumerate(zip(addr, reqnum)):
            send_error(addr, reqnum, 6, e.args[0])
        else:
          send_error(addr, reqnum, 6, e.args[0])
        continue
      status = int(0).to_bytes(8, 'little', signed=True)
      if batched:
        for i, (addr, reqnum) in enumerate(zip(addr, reqnum)):
          data = packlib.pack(packlib.tree_map(lambda x: x[i], result))
          outer.send(addr, reqnum, status, *data)
      else:
        data = packlib.pack(result)
        outer.send(addr, reqnum, status, *data)
    return waiting

  def send_error(addr, reqnum, status, message):
    assert 1 <= status, status
    status = status.to_bytes(8, 'little', signed=False)
    data = message.encode('utf-8')
    outer.send(addr, reqnum, status, data)
    if errors:
      raise RuntimeError(message)

  try:
    outer = server_socket.ServerSocket(outer_port, f'{name}Server', **kwargs)
    inner = client.Client(inner_port, f'{name}Client', **kwargs)
    batches = {}  # {method: ([addr], [reqnum], structure, [array])}
    jobs = []
    shutdown = False
    while running.is_set() or jobs:
      if running.is_set():
        maybe_recv(outer, inner, jobs, batches)
      elif not shutdown:
        shutdown = True
        outer.shutdown()
      jobs = maybe_send(outer, inner, jobs)
  finally:
    outer.close()
    inner.close()
